# -*- coding: utf-8 -*-
"""Top-level package for aiovault."""

from .client import VaultClient  # noqa: F401

__author__ = """Terry Cain"""
__email__ = 'terry@terrys-home.co.uk'
__version__ = '0.1.0'

__all__ = ['client', 'exceptions']
